<?php

namespace Drupal\salesforce_eca\Plugin\Action;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\eca\Plugin\Action\ConfigurableActionBase;
use Drupal\salesforce\Rest\RestClientInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Get Salesforce object metadata (describe).
 *
 * @Action(
 *   id = "salesforce_object_metadata",
 *   label = @Translation("Get Salesforce object metadata"),
 *   description = @Translation("Retrieve metadata about a Salesforce object type including field definitions, properties, and relationships."),
 *   eca_version_introduced = "1.0.0"
 * )
 */
class SalesforceObjectMetadata extends ConfigurableActionBase {

  /**
   * The Salesforce REST client.
   *
   * @var \Drupal\salesforce\Rest\RestClientInterface
   */
  protected RestClientInterface $salesforceClient;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    $instance = parent::create(
      $container,
      $configuration,
      $plugin_id,
      $plugin_definition
    );

    // Inject the Salesforce client service.
    $instance->setSalesforceClient($container->get('salesforce.client'));

    return $instance;
  }

  /**
   * Sets the Salesforce client.
   *
   * @param \Drupal\salesforce\Rest\RestClientInterface $salesforce_client
   *   The Salesforce REST client.
   */
  public function setSalesforceClient(RestClientInterface $salesforce_client): void {
    $this->salesforceClient = $salesforce_client;
  }

  /**
   * {@inheritdoc}
   */
  public function access($object, ?AccountInterface $account = NULL, $return_as_object = FALSE): bool|AccessResultInterface {
    // For ECA actions, we generally allow access since the action execution
    // is controlled by the ECA workflow configuration.
    $access_result = AccessResult::allowed();
    return $return_as_object ? $access_result : $access_result->isAllowed();
  }

  /**
   * {@inheritdoc}
   */
  public function execute($entity = NULL): void {
    // Process tokens in the object type.
    $object_type = $this->tokenService->replaceClear($this->configuration['object_type']);
    $result_token = $this->configuration['result_token'];

    try {
      // Log the metadata retrieval attempt.
      $this->logger->info('Retrieving Salesforce metadata for object type: @type', [
        '@type' => $object_type,
      ]);

      // Log filtering configuration.
      $filtering_enabled = [
        'picklist_values' => $this->configuration['include_picklist_values'],
        'address_fields' => $this->configuration['include_address_fields'],
        'geocode_fields' => $this->configuration['include_geocode_fields'],
        'system_fields' => $this->configuration['include_system_fields'],
        'relationship_fields' => $this->configuration['include_relationship_fields'],
      ];
      $this->logger->info('Metadata filtering configuration: @config', [
        '@config' => json_encode($filtering_enabled),
      ]);

      // Get the object metadata using objectDescribe.
      $describe = $this->salesforceClient->objectDescribe($object_type);

      // Extract field information into a more accessible format.
      $fields_by_name = [];
      $fields_list = [];
      $required_fields = [];
      $updateable_fields = [];
      $createable_fields = [];
      $relationship_fields = [];
      $picklist_fields = [];

      foreach ($describe->getFields() as $field_name => $field_info) {
        // Check if this field should be
        // excluded based on filtering configuration.
        if ($this->shouldExcludeField($field_name, $field_info)) {
          continue;
        }

        // Build simplified field data.
        $field_data = [
          'name' => $field_info['name'],
          'label' => $field_info['label'],
          'type' => $field_info['type'],
          'length' => $field_info['length'] ?? 0,
          'required' => !empty($field_info['nillable']) ? 0 : 1,
          'createable' => !empty($field_info['createable']) ? 1 : 0,
          'updateable' => !empty($field_info['updateable']) ? 1 : 0,
          'unique' => !empty($field_info['unique']) ? 1 : 0,
          'default_value' => $field_info['defaultValue'] ?? '',
          'help_text' => $field_info['inlineHelpText'] ?? '',
        ];

        // Add to main fields array.
        $fields_by_name[$field_name] = $field_data;
        $fields_list[] = $field_name;

        // Categorize fields.
        if (!empty($field_info['nillable']) === FALSE && empty($field_info['defaultValue'])) {
          $required_fields[] = $field_name;
        }
        if (!empty($field_info['updateable'])) {
          $updateable_fields[] = $field_name;
        }
        if (!empty($field_info['createable'])) {
          $createable_fields[] = $field_name;
        }
        if (!empty($field_info['referenceTo'])) {
          $relationship_fields[$field_name] = [
            'name' => $field_name,
            'references' => $field_info['referenceTo'],
            'relationship_name' => $field_info['relationshipName'] ?? '',
          ];
        }

        // Handle picklist fields - only include values if configured to do so.
        if ($field_info['type'] === 'picklist' && !empty($field_info['picklistValues']) && $this->configuration['include_picklist_values']) {
          $picklist_options = [];
          foreach ($field_info['picklistValues'] as $option) {
            if (!empty($option['active'])) {
              $picklist_options[$option['value']] = [
                'value' => $option['value'],
                'label' => $option['label'],
                'is_default' => !empty($option['defaultValue']) ? 1 : 0,
              ];
            }
          }
          $picklist_fields[$field_name] = $picklist_options;
        }
      }

      // Get record type information if available.
      $record_types = [];
      if (!empty($describe->recordTypeInfos)) {
        foreach ($describe->recordTypeInfos as $rt_info) {
          if (!empty($rt_info['available']) && !empty($rt_info['developerName'])) {
            $record_types[$rt_info['developerName']] = [
              'id' => $rt_info['recordTypeId'],
              'name' => $rt_info['name'],
              'developer_name' => $rt_info['developerName'],
              'is_default' => !empty($rt_info['defaultRecordTypeMapping']) ? 1 : 0,
            ];
          }
        }
      }

      // Build comprehensive token data structure.
      $token_data = [
        // Basic object information.
        'object_type' => $describe->getName(),
        'label' => $describe->label,
        'label_plural' => $describe->labelPlural,
        'key_prefix' => $describe->keyPrefix,

        // Object capabilities.
        'is_createable' => !empty($describe->createable) ? 1 : 0,
        'is_updateable' => !empty($describe->updateable) ? 1 : 0,
        'is_deletable' => !empty($describe->deletable) ? 1 : 0,
        'is_queryable' => !empty($describe->queryable) ? 1 : 0,
        'is_searchable' => !empty($describe->searchable) ? 1 : 0,
        'is_retrieveable' => !empty($describe->retrieveable) ? 1 : 0,
        'is_custom' => !empty($describe->custom) ? 1 : 0,

        // Field information.
        'fields' => $fields_by_name,
        'field_names' => $fields_list,
        'required_fields' => $required_fields,
        'updateable_fields' => $updateable_fields,
        'createable_fields' => $createable_fields,
        'relationship_fields' => $relationship_fields,
        'picklist_fields' => $picklist_fields,
        'field_count' => count($fields_list),

        // Record types.
        'record_types' => $record_types,
        'has_record_types' => !empty($record_types) ? 1 : 0,

        // URLs for various operations.
        'urls' => $describe->urls ?? [],

        // Success indicator.
        'success' => 1,
      ];

      // Store the token data.
      $this->tokenService->addTokenData($result_token, $token_data);

      // Calculate field count and filtering statistics.
      $total_fields_available = count($describe->getFields());
      $filtered_fields_count = count($fields_list);
      $excluded_fields_count = $total_fields_available - $filtered_fields_count;

      // Log success with filtering statistics.
      $this->logger->info('Successfully retrieved metadata for Salesforce object type @type. Included @included fields out of @total available (excluded @excluded fields based on filtering configuration).', [
        '@type' => $object_type,
        '@included' => $filtered_fields_count,
        '@total' => $total_fields_available,
        '@excluded' => $excluded_fields_count,
      ]);

    }
    catch (\Exception $e) {
      // Log the error.
      $this->logger->error('Error retrieving Salesforce object metadata: @error. Object type: @type', [
        '@error' => $e->getMessage(),
        '@type' => $object_type ?? 'unknown',
      ]);

      // Provide error information in token.
      $error_data = [
        'success' => 0,
        'error' => $e->getMessage(),
        'object_type' => $object_type ?? '',
        'fields' => [],
        'field_names' => [],
      ];

      $this->tokenService->addTokenData($result_token, $error_data);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'object_type' => '',
      'result_token' => 'salesforce_metadata',
      'include_picklist_values' => FALSE,
      'include_address_fields' => FALSE,
      'include_geocode_fields' => FALSE,
      'include_system_fields' => FALSE,
      'include_relationship_fields' => TRUE,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['object_type'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Salesforce object type'),
      '#description' => $this->t('The Salesforce object type to get metadata for (e.g., Contact, Account, Lead, CustomObject__c). You can use tokens in this field.'),
      '#default_value' => $this->configuration['object_type'],
      '#required' => TRUE,
    ];

    // Add performance options header.
    $form['performance_header'] = [
      '#type' => 'markup',
      '#markup' => '<h3>' . $this->t('Performance Options') . '</h3><p>' . $this->t('Configure what metadata to include. Excluding certain field types can significantly improve performance and reduce memory usage.') . '</p>',
    ];

    $form['include_picklist_values'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Include picklist values'),
      '#description' => $this->t('<strong>Default: Excluded</strong> - Include all possible values for picklist fields. <strong>Warning:</strong> This can significantly increase memory usage for objects with large picklists (e.g., country/state fields with hundreds of values).'),
      '#default_value' => $this->configuration['include_picklist_values'],
    ];

    $form['include_address_fields'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Include address fields'),
      '#description' => $this->t('<strong>Default: Excluded</strong> - Include address-related fields (Street, City, State, Country, PostalCode, Address compound fields). These often contain large picklists with hundreds of country/state values.'),
      '#default_value' => $this->configuration['include_address_fields'],
    ];

    $form['include_geocode_fields'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Include geocode fields'),
      '#description' => $this->t('<strong>Default: Excluded</strong> - Include geographic fields (Latitude, Longitude, GeocodeAccuracy). These are typically not needed for most workflows.'),
      '#default_value' => $this->configuration['include_geocode_fields'],
    ];

    $form['include_system_fields'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Include system fields'),
      '#description' => $this->t('<strong>Default: Excluded</strong> - Include system audit fields (CreatedDate, CreatedById, LastModifiedDate, LastModifiedById, SystemModstamp, etc.). Usually not needed for business workflows.'),
      '#default_value' => $this->configuration['include_system_fields'],
    ];

    $form['include_relationship_fields'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Include relationship fields'),
      '#description' => $this->t('<strong>Default: Included</strong> - Include lookup/relationship fields (AccountId, OwnerId, etc.). Usually needed for workflows.'),
      '#default_value' => $this->configuration['include_relationship_fields'],
    ];

    $form['result_token'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Result token name'),
      '#description' => $this->t('The token name to store the metadata results. You can access the metadata using:<br>
        <strong>Object information:</strong><br>
        • <code>[TOKEN:object_type]</code> - The object type name<br>
        • <code>[TOKEN:label]</code> - User-friendly label<br>
        • <code>[TOKEN:label_plural]</code> - Plural form of label<br>
        • <code>[TOKEN:key_prefix]</code> - 3-character ID prefix<br>
        <strong>Object capabilities:</strong><br>
        • <code>[TOKEN:is_createable]</code> - Can create records (1/0)<br>
        • <code>[TOKEN:is_updateable]</code> - Can update records (1/0)<br>
        • <code>[TOKEN:is_deletable]</code> - Can delete records (1/0)<br>
        • <code>[TOKEN:is_queryable]</code> - Can query records (1/0)<br>
        • <code>[TOKEN:is_custom]</code> - Is custom object (1/0)<br>
        <strong>Field information:</strong><br>
        • <code>[TOKEN:field_names]</code> - List of all field names<br>
        • <code>[TOKEN:field_count]</code> - Total number of fields<br>
        • <code>[TOKEN:required_fields]</code> - List of required field names<br>
        • <code>[TOKEN:updateable_fields]</code> - List of updateable field names<br>
        • <code>[TOKEN:createable_fields]</code> - List of createable field names<br>
        <strong>Field details (for any field):</strong><br>
        • <code>[TOKEN:fields:Email:label]</code> - Field label<br>
        • <code>[TOKEN:fields:Email:type]</code> - Field type (string, picklist, etc.)<br>
        • <code>[TOKEN:fields:Email:length]</code> - Maximum length<br>
        • <code>[TOKEN:fields:Email:required]</code> - Is required (1/0)<br>
        • <code>[TOKEN:fields:Email:updateable]</code> - Can be updated (1/0)<br>
        <strong>Picklist values:</strong><br>
        • <code>[TOKEN:picklist_fields:Status]</code> - All picklist options for Status field<br>
        • <code>[TOKEN:picklist_fields:Status:New:label]</code> - Label for specific option<br>
        <strong>Relationships:</strong><br>
        • <code>[TOKEN:relationship_fields:AccountId:references]</code> - What object it references<br>
        <strong>Record types:</strong><br>
        • <code>[TOKEN:record_types]</code> - All record types<br>
        • <code>[TOKEN:has_record_types]</code> - Has record types (1/0)<br>
        • <code>[TOKEN:record_types:DeveloperName:id]</code> - Record type ID<br>
        <strong>Status:</strong><br>
        • <code>[TOKEN:success]</code> - Operation success (1/0)<br>
        • <code>[TOKEN:error]</code> - Error message if failed'),
      '#default_value' => $this->configuration['result_token'],
      '#required' => TRUE,
      '#eca_token_reference' => TRUE,
      '#element_validate' => [[$this, 'validateTokenName']],
    ];

    return parent::buildConfigurationForm($form, $form_state);
  }

  /**
   * Validates the token name field.
   *
   * @param array $element
   *   The form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public function validateTokenName(array $element, FormStateInterface $form_state): void {
    $value = $element['#value'];
    if (!preg_match('/^[a-zA-Z_][a-zA-Z0-9_]*$/', $value)) {
      $form_state->setError($element, $this->t('Token name must start with a letter or underscore and contain only letters, numbers, and underscores.'));
    }
  }

  /**
   * Determines if a field should be excluded based on filtering configuration.
   *
   * @param string $field_name
   *   The field name.
   * @param array $field_info
   *   The field information from Salesforce.
   *
   * @return bool
   *   TRUE if the field should be excluded, FALSE otherwise.
   */
  private function shouldExcludeField(string $field_name, array $field_info): bool {
    $field_type = $field_info['type'] ?? '';

    // Address fields - exclude if not explicitly included.
    if (!$this->configuration['include_address_fields']) {
      $address_patterns = [
        '/^(Mailing|Other|Billing|Shipping)(Street|City|State|Country|PostalCode|Latitude|Longitude|GeocodeAccuracy|Address)$/i',
        '/^(Mailing|Other|Billing|Shipping)(StateCode|CountryCode)$/i',
      ];

      foreach ($address_patterns as $pattern) {
        if (preg_match($pattern, $field_name)) {
          return TRUE;
        }
      }

      // Address compound fields.
      if ($field_type === 'address') {
        return TRUE;
      }
    }

    // Geocode fields - exclude if not explicitly included.
    if (!$this->configuration['include_geocode_fields']) {
      $geocode_patterns = [
        '/Latitude$/i',
        '/Longitude$/i',
        '/GeocodeAccuracy$/i',
      ];

      foreach ($geocode_patterns as $pattern) {
        if (preg_match($pattern, $field_name)) {
          return TRUE;
        }
      }
    }

    // System fields - exclude if not explicitly included.
    if (!$this->configuration['include_system_fields']) {
      $system_fields = [
        'CreatedDate', 'CreatedById', 'LastModifiedDate', 'LastModifiedById',
        'SystemModstamp', 'LastViewedDate', 'LastReferencedDate',
        'LastActivityDate', 'LastCURequestDate', 'LastCUUpdateDate',
        'IsDeleted', 'MasterRecordId',
      ];

      if (in_array($field_name, $system_fields, TRUE)) {
        return TRUE;
      }
    }

    // Relationship fields - exclude if not explicitly included.
    if (!$this->configuration['include_relationship_fields']) {
      if (!empty($field_info['referenceTo']) || $field_type === 'reference') {
        return TRUE;
      }
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $this->configuration['object_type'] = $form_state->getValue('object_type');
    $this->configuration['result_token'] = $form_state->getValue('result_token');
    $this->configuration['include_picklist_values'] = $form_state->getValue('include_picklist_values');
    $this->configuration['include_address_fields'] = $form_state->getValue('include_address_fields');
    $this->configuration['include_geocode_fields'] = $form_state->getValue('include_geocode_fields');
    $this->configuration['include_system_fields'] = $form_state->getValue('include_system_fields');
    $this->configuration['include_relationship_fields'] = $form_state->getValue('include_relationship_fields');
    parent::submitConfigurationForm($form, $form_state);
  }

}
